#!/usr/bin/python
# -*- coding: latin-1 -*-

# External imports
from __future__ import with_statement
import ply.lex as lex

# Definition of class LexAnalysisError
class LexAnalysisError(Exception):
    def __init__(self, message, line, value):
        self.message = message
        self.line = line
        self.value = value
        
    def __repr__(self):
        return "LexAnalysisError(%s, %d, %s)" % (self.message, self.line, self.value)
        
    def __str__(self):
        return "%s at line %d (value %s)" % (self.message, self.line, self.value)
        
# List of reserved words
reserved_words = (
    'if',
    'else',
    'for',
    'while',
    'until',
    'true',
    'false'
)

# List of tokens
tokens = (
    'OP_UNARY',     # ++ --
    'OP_MUL',       # * / %
    'OP_ADD',       # + -
    'OP_RELA',      # < <= > >=
    'OP_COMP',      # == !=
    'AND_LOGIC',    # &&
    'OR_LOGIC',     # ||
    'OP_ASSIGN',    # = += -= *= /= %=
    'NUMBER',       # int / float / exponential
    'IDENTIFIER',
    'NEWLINE'
) + tuple(map(lambda s:s.upper(), reserved_words))

# List of litterals (token of one character)
literals = '.(){},![]'

# Definition of the token "OP_COMP"
# Support operators : == !=
def t_OP_COMP(t):
    r'[=!]='
    return t
    
# Definition of the token "OP_ASSIGN"
# Support operators : = *= /= += -= %=
def t_OP_ASSIGN(t):
    r'[*/+%-]?='
    return t
    
# Definition of the token "NUMBER"
# Support formats : int / float / exponential
def t_NUMBER(t):
    r'-?\d+(\.\d*)?(e-?\d+)?'
    try:
        t.value = float(t.value)   
        print t.value
    except ValueError:
        raise LexAnalysisError("Exception while parsing number", t.lineno, t.value)
    return t
    
# Definition of the token "OP_UNARY"
# Support operators : ++ --
def t_OP_UNARY(t):
    r'(\+\+)|(\-\-)'
    return t
    
# Definition of the token "OP_MUL"
# Support operators : * / %
def t_OP_MUL(t):
    r'[*/%]'
    return t
    
# Definition of the token "OP_ADD"
# Support operators : + -
def t_OP_ADD(t):
    r'[+-]'
    return t
    
# Definition of the token "OP_RELA"
# Support operators : < <= > >=
def t_OP_RELA(t):
    r'[<>]=?'
    return t
    
# Definition of the token "AND_LOGIC"
# Support operators : &&
def t_AND_LOGIC(t):
    r'&&'
    return t
    
# Definition of the token "OR_LOGIC"
# Support operators : ||
def t_OR_LOGIC(t):
    r'\|\|'
    return t
    
# Definition of the token "COMMENT"
# Support monoline comment : #
t_ignore_COMMENT = r'\#.*' # Comments are ignored

# Definition of the token "IDENTIFIER"
# Check for reserved word
def t_IDENTIFIER(t):
    r'[a-zA-Z_][a-zA-Z_0-9]*'
    if t.value in reserved_words:
        t.type = t.value.upper()
        t.value = ''
    return t
    
# Definition of the token "NEWLINE"
def t_NEWLINE(t):
    r'[\n\r]+'
    t.lexer.lineno += len(t.value.strip('\r'))
    t.value = ''
    return t
    
# Ignore white spaces
t_ignore  = ' \t'

# Catch error
def t_error(t):
    raise LexAnalysisError("Illegal character", t.lineno, repr(t.value[0]))
    
# Construct LEX
lex.lex()

# If the file is not used as a module
if __name__ == "__main__":
    import sys, os
    
    if len(sys.argv) == 2:
        # File open
        try:
            file = open(os.path.normpath(os.path.normcase(sys.argv[1])), 'r')
        except IOError:
            print "File error: Can't open the file '%s' in read mode!" % sys.argv[1]
            exit(-1)
    	    
    	# Read file using a context manager
        with file:
            prog = file.read()
            lex.input(prog)
            
            while 1:
                try:
                    tok = lex.token()
                except Exception, e:
                    print str(e)
                    exit(-1)
                    
                # No more token
                if not tok:
                    break
                    
                print "Line %d: %s(%s)" % (tok.lineno, tok.type, tok.value)
    else:
        print "Arguments error: Please specify a file as first argument!"
        exit(-1)